package jarnal;

import javax.swing.*;
import java.awt.*;
import java.awt.font.*;
import java.awt.geom.*;
import java.awt.event.*;
import java.awt.datatransfer.*;
import java.util.*;
import java.lang.Math.*;
import java.lang.Number.*;
import java.io.*;
import java.util.zip.*;
import javax.imageio.*;
import javax.imageio.stream.*;
import java.awt.image.BufferedImage;

import jarnal.Jtool;
import jarnal.Jpar;

public class Jpages{

	public static boolean highlightLines = false;

	public boolean saveBg = false;
	//public String bgStr = "";

	private Jpage current = new Jpage();
	private LinkedList jpages = new LinkedList();
	private LinkedList bgs = new LinkedList();
	private LinkedList redoList = new LinkedList();
	private LinkedList undoList = new LinkedList();
	private UndoPage undo;
	private int cindex = 0;
	private boolean savedUndo;
	private float scale = 1;
	private Jpaper defaultPaper = new Jpaper();
	private Jchar parms = new Jchar();
	private String bgfname = "";
	
	public Jpages(){
		jpages.add(current);
	}

	public int getHeight(){return current.getHeight();}
	public int getWidth(){return current.getWidth();}
	public int getBaseWidth(){return current.getBaseWidth();}
	public Rectangle drawLast(){return current.drawLast();}
	public Rectangle dragText(int x, int y){return current.dragText(x, y);}
	public String getPaperDesc(){return current.getPaper().getDesc();}
	public String getDefaultPaper(){return defaultPaper.getConf();}
	public void clipText(Toolkit toolkit){current.clipText(toolkit);}
	public String getDesc(){
		String desc = current.getDesc();
		if(desc == null) desc = parms.getDesc();
		return desc;
	}

	public void writeGraphicFile(File f, OutputStream os, String type){
	    int w = getWidth();
            int h = getHeight();
	    float scale = 1.0f;
	    BufferedImage g = new BufferedImage(w, h, BufferedImage.TYPE_3BYTE_BGR);
	    Graphics2D g2 = g.createGraphics();
            g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
            g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
            g2.setRenderingHint(RenderingHints.KEY_FRACTIONALMETRICS, RenderingHints.VALUE_FRACTIONALMETRICS_ON);	
            g2.setRenderingHint(RenderingHints.KEY_ALPHA_INTERPOLATION, RenderingHints.VALUE_ALPHA_INTERPOLATION_QUALITY);
	    g2.setBackground(Color.white);
	    g2.clearRect(0, 0, w, h);
	    if(print(g2, getPage() - 1, scale)) {
	    	try{
			if(f == null) ImageIO.write(g, type, os);
	    		else ImageIO.write(g, type, f);
	    	} catch(java.io.IOException test){System.out.println(test);}
	    }
	    else System.out.println("can't draw");
	} 

	public void setParms(String str){
		parms = Jpar.makeParms(str);
	}

	public String getParms(){
		return parms.getSVG();
	}

	public void setDefaultParms(){
		parms = current.getCurParms();
		if(parms == null) parms = new Jchar();
	}

	public void setDefaultPaper(String s){
		defaultPaper.setConf(s);
	}

	public void setWidth(float f){
		Jpaper pp = current.getPaper().copy();	
		pp.width = (int) (f * Jpaper.dpi);
		setPaper(pp);
	}	

	public void setHeight(float f){
		Jpaper pp = current.getPaper().copy();	
		pp.height = (int) (f * Jpaper.dpi);
		setPaper(pp);
	}

	public void toggleLandscape(){
		Jpaper pp = current.getPaper().copy();
		pp.height = current.getPaper().width;
		pp.width = current.getPaper().height;
		setPaper(pp);
	}

	public void newBg(){
		if(bgs.size() <= 1) return;
		if(cindex < bgs.size()) return;
		Jpaper pp = current.getPaper().copy();
		pp.showBg = 0;
		setPaper(pp);
	}

	public void toggleBackground(){
		Jpaper pp = current.getPaper().copy();
		if(current.getPaper().showBg == 1) pp.showBg = 0;
		else pp.showBg = 1;
		setPaper(pp);
	}

	public void setTransparency(int trans){
		Jpaper pp = current.getPaper().copy();
		pp.transparency = trans;
		setPaper(pp);
	}		

	public void setLines(int nlines){
		Jpaper pp = current.getPaper().copy();	
		pp.nlines = nlines;
		setPaper(pp);
	}

	public void setPaper(String p){
		Jpaper pp = current.getPaper().copy();	
		pp.paper = p;
		setPaper(pp);
	}

	private void setPaper(Jpaper p){
		undo = new UndoPage();
		undo.cindex = cindex;
		undo.op = "paper";
		undo.data = current.getPaper();
		undoList.add(undo);
		current.setPaper(p);
		defaultPaper = p.copy();
	}

	public float getScale(){
		return scale;
	}
	
	public void setScale(float scale){
		this.scale = scale;
		currentScale();
	}

	public void draw(Graphics2D g2){current.draw(g2);}
	public void draw(Graphics2D g2, int x, int y, int w, int h){current.draw(g2, x, y, w, h);}

	public boolean print(Graphics2D g2, int pi, float scale){
		if(pi >= jpages.size()) return false;
		Jpage jp = new Jpage();
		jp = (Jpage) jpages.get(pi);
		String txt = jp.save();
		jp = new Jpage();
		jp.open(txt);
		jp.print = true;
		jp.setScale(scale);
		if((bgs.size()) > 0 && (jp.getPaper().showBg == 1)){
			int bindex = pi;
			if(bindex >= bgs.size()) bindex = bgs.size() - 1;
			jp.bg = (BufferedImage) bgs.get(bindex);
		}			
		jp.draw(g2);
		return true;	
	}

	public void setSizeToBg(){
		if(bgs.size() == 0) {
			current.bg = null;
			return;
		}
		int bindex = cindex;
		if(bindex >= bgs.size()) bindex = bgs.size() - 1;
		if(current.getPaper().showBg == 1) {
			BufferedImage bi = (BufferedImage) bgs.get(bindex);
			Jpaper pp = current.getPaper().copy();	
			pp.height = bi.getHeight();
			pp.width = bi.getWidth();
			setPaper(pp);
		}
	}

	private void currentScale(){
		current.setScale(scale);
		if(bgs.size() == 0) {
			current.bg = null;
			return;
		}
		int bindex = cindex;
		if(bindex >= bgs.size()) bindex = bgs.size() - 1;
		if(current.getPaper().showBg == 1) current.bg = (BufferedImage) bgs.get(bindex);
		else current.bg = null;
	}

	public void upScale(int adj){
		if(adj == 1) scale = scale * 1.2f;
		if(adj == -1) scale = scale / 1.2f;
		if(adj == 0) scale = 1.0f;
		currentScale();
	}

	private UndoPage undoEntirePage(){
		undo = new UndoPage();
		undo.cindex = cindex;
		undo.op = "page";
		Jpage jp = new Jpage();
		jp.open(current.save());
		undo.data = jp;
		return undo;
	}

	public void clearPage(){
		undoList.add(undoEntirePage());
		Jpaper p = current.getPaper();
		current = new Jpage();
		jpages.set(cindex, current);
		current.setPaper(p);
	}

	public void replacePage(String str){
		undoList.add(undoEntirePage());
		Jpaper p = current.getPaper();
		current = new Jpage();
		jpages.set(cindex, current);
		if(!str.equals("")) current.open(str);
		else current.setPaper(p);
	}

	public void startDragOp(int op){
		undoList.add(undoEntirePage());
		current.startDragOp(op);
	}

	public void endDragOp(){current.endDragOp();}
	public void dragOp(Point2D.Double p){current.dragOp(p);}

	public void startStroke(Point2D.Double startL, Jtool jt){
		current.startStroke(startL, jt);
		undo = new UndoPage();
		if(jt.highlighter) undo.top = false;
		savedUndo = false;
	}

	public Rectangle startText(Point2D.Double startL){
		if(undoList.size() > 0){
			UndoPage undo = (UndoPage) undoList.getLast();
			undo.skiptext = false;
		}
		return current.startText(startL);
	}

	public void stroke(Point2D.Double endL){
		if(current.stroke(endL) && !savedUndo){
			undo.op = "stroke";
			undo.cindex = cindex;
			undoList.add(undo);
			savedUndo = true;
			redoList = new LinkedList();
		}
	}

	public Rectangle adv(int adv, int extend){
		return current.adv(adv, extend);
	}

	public Rectangle typeKey(String str, Jtool jt){
		undo = new UndoPage();
		String oldXML = current.typeKey(str, jt, parms);
		undo.oindex = current.getText();
		if(oldXML == null) undo.op = "newText";
		else {
			undo.op = "typeKey";
			undo.data = oldXML;
		}
		undo.skiptext = true;
		undo.cindex = cindex;
		if((str.length() == 1) && (!(str.equals(" ") || str.equals("\n")))){
			//undo.skiptext = true;
			if(undoList.size() > 0){
				UndoPage topUndo = (UndoPage) undoList.getLast();
				if((!topUndo.skiptext) || (topUndo.cindex != undo.cindex) ||(topUndo.oindex != undo.oindex)) undoList.add(undo);
			}
			else undoList.add(undo);
		}
		else undoList.add(undo);
		redoList = new LinkedList();
		return current.getTextRectangle();
	}

	public Rectangle setSelStyle(boolean bold, boolean italic, boolean underline, Float size, String font, String color){
		undo = new UndoPage();
		String oldXML = current.setSelStyle(bold, italic, underline, size, font, color, parms);
		undo.oindex = current.getText();
		if(oldXML == null) undo.op = "newText";
		else {
			undo.op = "typeKey";
			undo.data = oldXML;
		}
		undo.cindex = cindex;
		undoList.add(undo);
		redoList = new LinkedList();
		return current.getTextRectangle();
	}


	public void click(Point2D.Double endL, Jtool jt){
		current.click(endL, jt);
		undo = new UndoPage();
		if(jt.highlighter) undo.top = false;
		undo.op = "stroke";
		undo.cindex = cindex;
		undoList.add(undo);
		redoList = new LinkedList();
	}


	public boolean undo(){
		boolean repage = true;
		if(undoList.size() == 0) return false;
		undo = (UndoPage) undoList.removeLast();
		if(cindex == undo.cindex) repage = false;
		cindex = undo.cindex;
		current = (Jpage) jpages.get(cindex);
		if (undo.op.equals("paper")){
			Jpaper p = current.getPaper();
			current.setPaper((Jpaper) undo.data);
			undo.data = p;
			redoList.add(undo);
			repage = true;
		}
		else if (undo.op.equals("stroke")){
			currentScale();
			undo.data = current.undo(undo.top);
			redoList.add(undo);
		}
		else if (undo.op.equals("newText")){
			currentScale();
			current.unselectText();
			undo.data = current.undo(undo.top);
			redoList.add(undo);
		}
		else if (undo.op.equals("typeKey")){
			String oldXML = (String) undo.data;
			undo.data = current.undoText(undo.oindex, oldXML);
			redoList.add(undo);
		}
		else if(undo.op.equals("pagebefore")){
			undo.op = "pagedelete";
			undo.data = jpages.remove(cindex);
			redoList.add(undo);
			if(cindex >= jpages.size()) cindex = jpages.size() - 1;
			current=(Jpage) jpages.get(cindex);
			repage = true;
		}
		else if(undo.op.equals("pagedelete")){
			current = (Jpage) undo.data;
			currentScale();
			undo.op = "pagebefore";
			undo.data = null;
			redoList.add(undo);
			jpages.add(cindex, current);
			repage = true;
		}
		else if(undo.op.equals("page")){
			current = (Jpage) jpages.remove(cindex);
			jpages.add(cindex, (Jpage) undo.data);
			undo.data = current;
			redoList.add(undo);
			current = (Jpage) jpages.get(cindex);
			repage = true;
		}
		else if(undo.op.equals("bg")){
			String bgtemp = (String) undo.data;
			undo.data = bgfname;
			redoList.add(undo);
			openBgx(bgtemp);
		}	
		currentScale();
		return repage;	
	}

	public boolean redo(){
		boolean repage = true;
		if(redoList.size() == 0) return false;
		undo = (UndoPage) redoList.removeLast();
		if(cindex == undo.cindex) repage = false;
		cindex = undo.cindex;
		current = (Jpage) jpages.get(cindex);
		if (undo.op.equals("paper")){
			Jpaper p = current.getPaper();
			current.setPaper((Jpaper) undo.data);
			undo.data = p;
			undoList.add(undo);
			repage = true;
		}
		else if (undo.op.equals("stroke")){
			currentScale();
			current.redo(undo);
			undoList.add(undo);
		}
		else if (undo.op.equals("newText")){
			currentScale();
			current.unselectText();
			current.redo(undo);
			undoList.add(undo);
		}
		else if (undo.op.equals("typeKey")){
			String oldXML = (String) undo.data;
			undo.data = current.undoText(undo.oindex, oldXML);
			undoList.add(undo);
		}
		else if(undo.op.equals("pagebefore")){
			undo.op = "pagedelete";
			undo.data = jpages.remove(cindex);
			undoList.add(undo);
			if(cindex >= jpages.size()) cindex = jpages.size() - 1;
			current=(Jpage) jpages.get(cindex);
			repage = true;
		}
		else if(undo.op.equals("pagedelete")){
			current = (Jpage) undo.data;
			currentScale();
			undo.op = "pagebefore";
			undo.data = null;
			undoList.add(undo);
			jpages.add(cindex, current);
			repage = true;
		}
		else if(undo.op.equals("page")){
			current = (Jpage) jpages.remove(cindex);
			jpages.add(cindex, (Jpage) undo.data);
			undo.data = current;
			undoList.add(undo);
			current = (Jpage) jpages.get(cindex);
			repage = true;
		}
		else if(undo.op.equals("bg")){
			String bgtemp = (String) undo.data;
			undo.data = bgfname;
			undoList.add(undo);
			openBgx(bgtemp);
		}
		currentScale();
		return repage;
	}

	public boolean nextPage(int delta){
		boolean answer = false;
		cindex = cindex + delta;
		if(cindex < 0) cindex = 0;
		if(cindex >= jpages.size()){
			answer = true;
			cindex = jpages.size() - 1;
		}
		current = (Jpage) jpages.get(cindex);
		currentScale();
		return answer;
	}

	public void pageDup(){
		pageAfter(current.save());
	}

	public void pageAfter(String str){
		cindex = cindex + 1;
		pageBefore(str);
	}

	public void pageBefore(String str){
		current = new Jpage();
		current.setPaper(defaultPaper);
		currentScale();
		undo = new UndoPage();
		undo.cindex = cindex;
		undo.op = "pagebefore";
		undoList.add(undo);
		redoList = new LinkedList();
		jpages.add(cindex, current);
		if(!str.equals("")) current.open(str);
	}

	public void pageDelete(){
		if(jpages.size() == 1) return;
		undo = new UndoPage();
		undo.cindex = cindex;
		undo.op = "pagedelete";
		undo.data = jpages.remove(cindex);
		undoList.add(undo);
		redoList = new LinkedList();
		if(cindex >= jpages.size()) cindex = jpages.size() - 1;
		current=(Jpage) jpages.get(cindex);
		currentScale();		
	}

	public int getPage(){
		return cindex + 1;
	}

	public int getPages(){
		return jpages.size();
	}

	public JarnalSelection copyPage(){
		return new JarnalSelection(current.getAllText(), current.save(), "");
	}

	public void netWrite(OutputStream out, String op, String conf){
		if(op.equals("$$jarnal")){
			save(out, conf);
		}
		if(op.equals("$$snapshot")){
			writeGraphicFile(null, out, "jpg");
		}
	}

	public void save(String fname, String conf){
		try{
			FileOutputStream out = new FileOutputStream(fname);
			save(out, conf);
		}
		catch(IOException ex){};
	}
	
	public void save(OutputStream out, String conf){
		String str;
		try{
			
			ZipOutputStream zip = new ZipOutputStream(out);
			Jpage outp;
			ZipEntry subname = new ZipEntry("jarnal.conf");
			zip.putNextEntry(subname);
			zip.write(conf.getBytes());
			if(saveBg && (bgfname != "")){
				subname = new ZipEntry("background");
				zip.putNextEntry(subname);
				FileInputStream in3 = new FileInputStream(bgfname);
				int nmax = 1000000; 
				byte b[] = new byte[nmax];
				int nread = 0;
				String bgStr = "";
				while((nread = in3.read(b)) >= 0){
					bgStr = bgStr + new String(b, 0, nread);
				}
				zip.write(bgStr.getBytes());
			}
			for(int i = 0; i < jpages.size(); i++){
				subname = new ZipEntry("p" + i + ".svg");
				zip.putNextEntry(subname);
				outp = (Jpage) jpages.get(i);
				str = outp.save();
				zip.write(str.getBytes());
			}
			zip.close();
		}
		catch(IOException ex){System.out.println(ex);};
	}

	public void undoBg(){
		undo = new UndoPage();
		undo.cindex = cindex;
		undo.op = "bg";
		undo.data = bgfname;
		undoList.add(undo);
		redoList = new LinkedList();
	}

	public void openBg(String fname){
		undoBg();
		openBgx(fname);
	}

	public void openBgx(String fname){
		bgfname = fname;
		bgs = new LinkedList();
		if(fname.equals("")) return;
		try{
			FileInputStream in1 = new FileInputStream(fname);
			FileInputStream in2 = new FileInputStream(fname);
			openBgx(in1, in2);
			//FileInputStream in3 = new FileInputStream(fname);
			//int nmax = 1000000; 
			//byte b[] = new byte[nmax];
			//int nread = 0;
			//bgStr = "";
			//while((nread = in3.read(b)) >= 0){
			//	bgStr = bgStr + new String(b, 0, nread);
			//}
		}
		catch(IOException ex){System.out.println(ex);}
	}

	public void openBgx(InputStream in1, InputStream in2){
		try{
			ZipInputStream zip = new ZipInputStream(in1);
			BufferedImage bg;
			int iind = 0;
			ImageIO.scanForPlugins();
			for(ZipEntry ze = zip.getNextEntry(); ze != null; ze = zip.getNextEntry()){
				bg = ImageIO.read(zip);
				bgs.add(bg);
				if(iind < jpages.size()) ((Jpage) jpages.get(iind)).bg = bg;
				iind++;	
			}
			if(iind == 0){
				//String test[] = ImageIO.getReaderFormatNames();
				//for(int ix = 0; ix < test.length; ix++) System.out.println(test[ix]);
				ImageInputStream iis = ImageIO.createImageInputStream (in2);
				Iterator readers = ImageIO.getImageReaders(iis);
                        	ImageReader reader = (ImageReader)readers.next();
				reader.setInput (iis, false);
				ImageReadParam ip = reader.getDefaultReadParam();
                	        int np = reader.getNumImages(true);
 				for (int i = 0; i < np; i++){
                                	bg = reader.read (i, ip);
					if(bg != null){
						bgs.add(bg);
						if(iind < jpages.size()) ((Jpage) jpages.get(iind)).bg = bg;
						iind++;
					} 
                        	}
			}	
		}
		catch(IOException ex){System.out.println(ex);}
	}

	public void removeBg(){
		undoBg();
		bgs = new LinkedList();
		currentScale();
	}
	
	public String open(String fname){
		String conf = "";
		jpages = new LinkedList();
		try{
			FileInputStream in = new FileInputStream(fname);
			ZipInputStream zip = new ZipInputStream(in);
			for(ZipEntry ze = zip.getNextEntry(); ze != null; ze = zip.getNextEntry()){
				int nmin = 1000000;
				int nborg = 40000;
				int nmax = nmin + ( 5 * nborg); 
				byte b[] = new byte[nmax];
				int nread = 0;
				String instr = "";
				int noff = 0;

				while((nread = zip.read(b, noff, nborg)) >= 0){
					noff = noff + nread;
					if(noff > nmax - (2 * nborg)){
						instr = instr + new String(b, 0, noff);
						noff = 0;
					}
				}
				instr = instr + new String(b, 0, noff);
				if(ze.getName().equals("jarnal.conf")) conf = instr;
				else {
					if(ze.getName().equals("background")){
						byte c[] = instr.getBytes();
						if(c.length > 0){
							File tfile = File.createTempFile("jarnalBg", ".tmp");
							bgfname = tfile.getPath();
							System.out.println(bgfname);
							//tfile.deleteOnExit();
							FileOutputStream tout = new FileOutputStream(tfile);
							tout.write(c);
							openBg(bgfname);
							//InputStream in1 = new ByteArrayInputStream(c);
							//InputStream in2 = new ByteArrayInputStream(c);
							//bgs = new LinkedList();
							//openBgx(in1, in2);
						 	//bgStr = new String(instr);	
						}	
					}
					else {
						current = new Jpage();
						jpages.add(current);
						current.open(instr);
					}
				}
			}		
		}
		catch(IOException ex){jpages.add(new Jpage());}	
		current = (Jpage) jpages.get(0);
		return conf;
	}
}

class Jpaper{
	public static int dpi = 84;
	public static int standardHeight = 861;
	public int height = 861;
	public int width = 714;
	public int nlines = 25;
	public String paper = "Lined";
	public int showBg = 1;
	public int transparency = 255;
	public String getConf(){
		return "paper=" + paper + "\nlines=" + nlines + "\nheight=" + height + "\nwidth=" + width + "\nbg=" + showBg + "\ntransparency=" + transparency + "\n";
	}
	public String getDesc(){return "" + ((float)width/dpi) + "\" X " + ((float)height/dpi) + "\"";}
	public void setConf(String s){
		String z = Jtool.getLine(s, "paper");
		if(z != null) paper = z;
		z = Jtool.getLine(s, "lines");
		if(z != null) nlines = Integer.parseInt(z);
		z = Jtool.getLine(s, "height");
		if(z != null) height = Integer.parseInt(z);
		z = Jtool .getLine(s, "width");
		if(z != null) width = Integer.parseInt(z);
		z = Jtool.getLine(s, "bg");
		if(z != null) showBg = Integer.parseInt(z);
		z = Jtool.getLine(s, "transparency");
		if(z != null) transparency = Integer.parseInt(z);
	}
	public Jpaper copy(){
		Jpaper p = new Jpaper();
		p.height = height;
		p.width = width;
		p.nlines = nlines;
		p.paper = paper;
		p.showBg = showBg;
		p.transparency = transparency;
		return p;
	}
}

class Jpage{

	static String header = "<?xml version=\"1.0\" standalone=\"no\"?>\n<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 20010904//EN\" \"http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd\">\n";
	static String middle = "xmlns=\"http://www.w3.org/2000/svg\">\n<title>Jarnal document - see http://www.dklevine/general/software/tc1000/jarnal.htm for details</title>\n";
	static String footer = "</svg>";

	private LinkedList strokes = new LinkedList();
	private LinkedList dragList;
	int dragOp = 0;
	private Point2D.Double dragStart;
	private Point2D.Double drag;
	private Point2D.Double startL;
	private Jstroke current;
	private int itext = -1;
	private Point2D.Double ptext;
	private Point2D.Double ptext2 = null;
	private float textPadding;
	private boolean dragCorner;

	private Jpaper ppr = new Jpaper();
	public  BufferedImage bg = null;

	private float scale = 1;
	private Jtool jtool;
	public boolean print = false;
	Graphics2D gg;
	private Line2D.Double border;
	BasicStroke bs;

	public Jpage(){
		strokes.add(new Jstroke(this));
		float x = (ppr.standardHeight * scale) / ppr.nlines;
		ptext = new Point2D.Double(2.2 * x, 3 * x);
	}

	private Jtext ctext(){
		if(itext == -1) return null;
		return (Jtext) strokes.get(itext);
	}

	private boolean highlight;

	public int getHeight(){ return (int) (ppr.height * scale);}
	public int getWidth(){ return (int) (ppr.width * scale);}
	public int getBaseWidth(){ return ppr.width;}
	public int getBaseHeight(){return ppr.height;}
	public float getScale(){return scale;}
	public void clipText(Toolkit toolkit){ if(ctext() != null) ctext().clip(toolkit);}
	
	private String getHeader(){
		return header + "<svg width=\"" + getBaseWidth() + "px\" height=\"" + getBaseHeight() + "px\" " + middle + "<desc>\n[Jarnal Page Parameters]\n" + ppr.getConf() + "</desc>\n";
	}

	public void setScale(float scale){
		this.scale = scale;
	}

	public void setPaper(Jpaper p){
		ppr = p;
	}

	public Jpaper getPaper(){
		return ppr;
	}

	public int getTransparency(){
		return ppr.transparency;
	}

	public String getAllText(){
		String all = "";
		LinkedList ll = new LinkedList();
		for(int i = 0; i < strokes.size(); i++){
			Jstroke js = (Jstroke) strokes.get(i);
			if(js.isText) {
				Jtext jt = (Jtext) js;
				boolean now = false;
				int j = 0;
				for(j = 0; j < ll.size(); j++){
					Jtext jj = (Jtext) ll.get(j);
					if(jt.corner.getY() < jj.corner.getY()) now = true;
					if((jt.corner.getY() == jj.corner.getY()) && (jt.corner.getX() < jj.corner.getX())) now = true;
					if(now) break;
				}
				ll.add(j, jt);
			}
		}
		for(int j = 0; j < ll.size(); j++) all = all + ((Jtext)ll.get(j)).getText() + "\n\n";
		return all;
	}

	public Jstroke undo(boolean top){
		if(strokes.size() == 0) return null;
		Jstroke swap;
		if(top) swap = (Jstroke) strokes.removeLast();
		else swap = (Jstroke) strokes.remove(0);
		return swap;
	}

	public void redo(UndoPage undo){
		Jstroke swap = (Jstroke) undo.data;
		if(undo.top) strokes.add(swap);
		else strokes.add(0, swap);
	}

	public int getText(){
		return itext;
	}

	public String getDesc(){
		if(itext > -1) return ctext().getDesc();
		return null;
	}

	public Jchar getCurParms(){
		if(itext > -1) return ctext().getCurParms();
		return null;
	}

	public void unselectText(){
		itext = -1;
	}

	public void endDragOp(){
		if((dragOp == 4) || (dragOp == 5)){
			for (Iterator i = strokes.iterator(); i.hasNext(); ){
				Jstroke js = (Jstroke) i.next();
				if(dragOp == 4){
					if(!js.below((int) dragStart.getY())) i.remove();
				}
				if(dragOp == 5){
					if(js.below((int) dragStart.getY())) i.remove();
				}
			}
		}
		dragOp = 0;
	}

	public void startDragOp(int op){
		dragOp = op;
		dragStart = startL;
		dragList = new LinkedList();
		for (Iterator i = strokes.iterator(); i.hasNext(); ){
		//for(int i = 0; i < strokes.size(); i++){
			//Jstroke js = (Jstroke) strokes.get(i);
			Jstroke js = (Jstroke) i.next();
			if(dragOp == 1){
				if(js.below((int) dragStart.getY())) dragList.add(js);
			}
			if(dragOp == 2){
				if(js.hit(dragStart)) dragList.add(js);
			}
		}
	}

	public void dragOp(Point2D.Double p){
		Point2D.Double sve = p;
		if(dragOp < 3){
			if(dragOp == 1) p = new Point2D.Double(0, p.getY() - dragStart.getY());
			else p = new Point2D.Double(p.getX() - dragStart.getX(), p.getY() - dragStart.getY()); 
			for(int i = 0; i < dragList.size(); i++){
				Jstroke js = (Jstroke) dragList.get(i);
				js.offset(p);
			} 
		}
		if(dragOp == 3){
			itext = -1;
			for (Iterator i = strokes.iterator(); i.hasNext(); ){
				Jstroke jstroke = (Jstroke) i.next(); 
				if(jstroke.hit(dragStart)) i.remove();
			}	
		}
		dragStart = sve;
	}
	
	public void startStroke(Point2D.Double startL, Jtool jt){
		this.startL = startL;
		current = null;
		jtool = jt;	
	} 

	public Rectangle dragText(int x, int y){
		if(ctext() == null) return null;
		if(!dragCorner) return ctext().dragText(x, y, scale);
		return ctext().setWidth((int)((float) x / scale));
	}

	public Rectangle startText(Point2D.Double startL){
		dragCorner = false;
		if(ptext2 != null){
			boolean isCorner = true;
			double x = scale * ptext2.getX();
			double y = scale * ptext2.getY();
			double z = (double) (textPadding/scale);
			if(startL.getX() < x - z) isCorner = false;
			if(startL.getX() > x + z) isCorner = false;
			if(startL.getY() < y - z) isCorner = false;
			if(startL.getY() > y + z) isCorner = false;
			if(isCorner){
				dragCorner = true;
				return null;
			}
		}
		for(int i = 0; i < strokes.size(); i++){
			Jstroke jstroke = (Jstroke) strokes.get(i);
			if(jstroke.isText){
				Jtext jt = (Jtext) jstroke;
				if(jt.hit(startL)){
					if(itext == i){
						ptext = jt.hitSel(startL, scale);
						return jt.getTextRectangle();
					}
					if(ctext() != null) {
						ctext().clearSel();
					}
					itext = i;
					ptext = jt.hitSel(startL, scale);
					return null;	
				}
			}
		}
		if(ctext() != null) {
			ctext().clearSel();
		}
		itext = -1;
		//ptext = startL;
		ptext = new Point2D.Double(startL.getX()/scale,startL.getY()/scale);
		return null;
	}

	public boolean stroke(Point2D.Double endL){
		if(startL != null){
			if(current == null){
				current = new Jstroke(this, jtool);
				if(!jtool.highlighter) strokes.add(current);
				else {
					strokes.add(0,current);
					current.isHighlight = true;
				}
			}	
			current.add(startL, endL);
			startL = endL;
			return true;
		}
		return false;
	}

	public Rectangle adv(int adv, int extend){
		return ctext().adv(adv, extend);
	}

	public String typeKey(String str, Jtool jt, Jchar parms){
		if(ctext() == null){
			Point2D.Double p = ptext;
			int cwidth = (int) (ppr.width - p.getX() - (2 * ppr.standardHeight/ppr.nlines));
			strokes.add(new Jtext(this, p, jt, cwidth, parms));
			itext = strokes.size() - 1;
			ctext().typeKey(str);
			return null;
		}
		return ctext().typeKey(str);
	}

	public String setSelStyle(boolean bold, boolean italic, boolean underline, Float size, String font, String color, Jchar parms){
		if(ctext() == null){
			Point2D.Double p = ptext;
			int cwidth = (int) (ppr.width - p.getX() - (2 * ppr.standardHeight/ppr.nlines));
			strokes.add(new Jtext(this, p, new Jtool(), cwidth, parms));
			itext = strokes.size() - 1;
			ctext().setSelStyle(bold, italic, underline, size, font, color);;
			return null;
		}
		return ctext().setSelStyle(bold, italic, underline, size, font, color); 
	}


	public String undoText(int oindex, String str){
		itext = oindex;
		if(ctext() == null) return "";
		String oldXML = ctext().save();
		ctext().make(str);
		return oldXML;
	}

	public void click(Point2D.Double endL, Jtool jt){
		if(!jt.highlighter) strokes.add(new Jstroke(this, jt, endL));
		else {
			Jstroke jse = new Jstroke(this, jt, endL);
			strokes.add(0, jse);
			jse.isHighlight = true;
		}
	}

	private Rectangle fixR(Rectangle r){
		int z = (int) textPadding;
		return new Rectangle((int) r.getX() - z, (int) r.getY() - z, r.width + (2*z), r.height + (2 *z));
	}

	private void drawPaper(Graphics2D g2){
		float swidth = 0.5f;
		bs = new BasicStroke(swidth);
		g2.setStroke(bs);

		float x = (ppr.standardHeight * scale) / ppr.nlines;
		float y = 2 * x;
		textPadding = x/5;

		if(ppr.paper.equals("Lined")){
			Point2D.Double lmu = new Point2D.Double((double) y, 0.0);
			border = new Line2D.Double((double) y, 0.0 , (double) y, (double)getHeight());
			g2.setPaint(Color.orange);
			g2.draw(border);
			border = new Line2D.Double((double) (y + x/4), 0.0 , (double) (y + x/4), (double)getHeight());
			g2.draw(border);
		}
		if(ppr.paper.equals("Lined") || ppr.paper.equals("Graph")){
			int start = 3;
			if(ppr.paper.equals("Graph")) start = 1;
			g2.setPaint(Color.cyan);
			int i = start;
			while((i * x) < getHeight()){
				border = new Line2D.Double(0.0, i * x, (double) getWidth(), i * x);
				g2.draw(border);
				i++;
			}
		}
		if(ppr.paper.equals("Graph")){
			g2.setPaint(Color.cyan);
			int i = 1;
			while((i * x) < getWidth()){
				border = new Line2D.Double(i * x, 0.0, i * x,(double) getHeight());
				g2.draw(border);
				i++;
			}
		}

		if(!print){
			g2.setPaint(Color.gray);
			Point2D.Double lu = new Point2D.Double(0.0, 0.0);
			Point2D.Double lb = new Point2D.Double(0.0, (double)(getHeight() - 1));
			Point2D.Double ru = new Point2D.Double((double)(getWidth() - 1), 0.0);
			Point2D.Double rb = new Point2D.Double((double)(getWidth() - 1), (double)(getHeight() - 1));
			border = new Line2D.Double(lu, ru);
			g2.draw(border);
			border = new Line2D.Double(lu, lb);
			g2.draw(border);
			border = new Line2D.Double(ru, rb);
			g2.draw(border);
			border = new Line2D.Double(lb, rb);
			g2.draw(border);
		}
		if((dragOp == 1) || (dragOp == 4) || (dragOp == 5)) border = new Line2D.Double(0.0, dragStart.getY(), (double)(getWidth() -1), dragStart.getY());
	}

	private void drawTextMark(Graphics2D g2, Point2D.Double pt){
		float swidth = 0.5f;
		bs = new BasicStroke(swidth);
		g2.setStroke(bs);
		g2.setPaint(Color.gray);
		float z = textPadding;
		Point2D.Double p = new Point2D.Double(pt.getX() * scale, pt.getY() * scale);
		Point2D.Double first = new Point2D.Double(p.getX() - z, p.getY());
		Point2D.Double second = new Point2D.Double(p.getX() + z, p.getY());
		Line2D.Double border = new Line2D.Double(first, second);
		g2.draw(border);	
		first = new Point2D.Double(p.getX(), p.getY() - z);
		second = new Point2D.Double(p.getX(), p.getY() + z);
		border = new Line2D.Double(first, second);
		g2.draw(border);
	}

	public Rectangle drawLast(){
		if(gg == null) return new Rectangle(0,0,1,1);
		Jstroke jstroke = (Jstroke) strokes.getLast();
		jstroke.draw(gg, print);
		return jstroke.getRectangle();
	}

	public Rectangle getTextRectangle(){
		if(gg == null) return new Rectangle(0,0,1,1);
		return ctext().getTextRectangle();
	}

	private void drawTextMarks(Graphics2D g2){
		if(!print){
			drawTextMark(g2, ptext);
			if(ctext() != null){
				ptext2 = new Point2D.Double(ptext.getX() + ctext().getWidth(), ptext.getY());
				drawTextMark(g2, ptext2);
			}
			else ptext2 = null;
		}
	}

	public void drawBg(Graphics2D g2){
		if(bg == null) return;
		AffineTransform at = new AffineTransform();
		at.scale(scale, scale);
		g2.drawImage(bg, at, null);
	}

	public void draw(Graphics2D g2, int x, int y, int w, int h){
		gg = g2;
		drawBg(g2);
		if(!Jpages.highlightLines) drawPaper(g2);
		for (Iterator i = strokes.iterator(); i.hasNext(); ){
			Jstroke jstroke = (Jstroke) i.next();
	    		if(jstroke.overlaps(x, y, w, h)) jstroke.draw(g2, print);
			if(jstroke.isPage && Jpages.highlightLines) drawPaper(g2);
		}
		drawTextMarks(g2);
	}	

	public void draw(Graphics2D g2){
		gg = g2;
		drawBg(g2);
		if(!Jpages.highlightLines) drawPaper(g2);
		for (Iterator i = strokes.iterator(); i.hasNext(); ){
			Jstroke jstroke = (Jstroke) i.next();
	    		if(jstroke.draw(g2, print) && Jpages.highlightLines) drawPaper(g2);
		}
		drawTextMarks(g2);
		if((dragOp == 1) || (dragOp == 4) || (dragOp == 5))	{
			g2.setStroke(bs);
			g2.setPaint(Color.red);
			g2.draw(border);
		}		
	}

	public String save(){
		String str = getHeader();
		for (Iterator i = strokes.iterator(); i.hasNext(); ){
			Jstroke jstroke = (Jstroke) i.next();
			str = str + jstroke.save();	
		}
		str = str + footer;
		return str;
	}

	public void open(String str){
		highlight = true;
		int pos = str.indexOf("<desc>");
		String strstr = str.substring(pos);
		if(pos >= 0){
			pos = strstr.indexOf("</desc>");
			if(pos >= 0) {
				strstr = strstr.substring(0,pos) + "\n";
				ppr.setConf(strstr);
			}
		}
		strokes = new LinkedList();
		pos = str.indexOf("\n\n\n");
		if(pos > 0) {
			strstr = str.substring(0, pos);
			str = str.substring(pos + 1);
			openStrokes(strstr);
		}
		strokes.add(new Jstroke(this));
		highlight = false;
		openStrokes(str);
	}

	private void openStrokes(String str){
		int pos = 0;
		int posc = 0;
		scale = 1.0f;
		boolean circle;
		boolean text;
		String strstr;
		while(pos >=0){
			circle = false;
			text = false;
			pos = str.indexOf("<path");
			posc = str.indexOf("<circle");
			if((pos == -1) || ((posc < pos) && (posc != -1))) {
				pos = posc;
				circle = true;
			}
			posc = str.indexOf("<text");
			if((pos == -1) || ((posc < pos) && (posc != -1))){
				pos = posc;
				text = true;
			}
			if(pos < 0) return;
			str = str.substring(pos);
			if(!text){
				int last = str.indexOf("/>");
				if(last < 0) return;
				strstr = str.substring(0, last);
				str = str.substring(last + 2);
				current = new Jstroke(this, strstr, circle);
				current.isHighlight = highlight;
			}
			else {
				int last = str.indexOf("</text>");
				if(last < 0) return;
				strstr = str.substring(0, last);
				str = str.substring(last + 6);				
				current = new Jtext(this, strstr);
			}
			strokes.add(current);
		}
	}
}

class UndoPage{
	public int cindex;
	public boolean top = true;
	public String op;
	public Object data;
	public int oindex;
	public boolean skiptext = false;
}

class Jstroke{

	static private float cwidth = 1.8f;

	private LinkedList lines = new LinkedList();
	private Point2D.Double click = null;
	public Jpage jpage;
	public Jtool jtool = new Jtool();
	public int xmin;
	public int xmax;
	public int ymin;
	public int ymax;
	public boolean isPage = false;
	public boolean isText = false;
	public boolean isHighlight = false;

	public Jstroke(){}

	public Jstroke(Jpage parent){
		isPage = true;
	}

	public Jstroke(Jpage parent, Jtool jt){
		jpage = parent;
		jtool.copy(jt);
	}

	public Jstroke(Jpage parent, Jtool jt, Point2D.Double click){
		jpage = parent;
		setClick(parent, jt, click);
	}

	public Jstroke(Jpage parent, String str, boolean circle){
		jpage = parent;
		if(circle){
			String a = getParm(str, "cx=");
			if (a == null) return;
			int x = Integer.parseInt(a, 10);
			a = getParm(str, "cy=");
			if (a == null) return;
			int y = Integer.parseInt(a, 10);
			a = getParm(str, "r=");
			if(a == null) return;
			float r = Float.parseFloat(a);
			a = getParm(str, "fill=");
			if(a == null) return;
			jtool.width = r;
			jtool.color = a.trim();
			click = new Point2D.Double((double) x, (double) y);
			updateRectangle();
			return;
		}
		String stroke = getParm(str, "stroke=");
		if(stroke == null) return;
		String b = getParm(str, "stroke-width=");
		if(b == null) return;
		float s = Float.parseFloat(b);
		jtool.width = s;
		jtool.color = stroke;		
		int pos = str.indexOf("\"");
		if (pos < 0) return;
		str = str.substring(pos + 1);
		pos = str.indexOf("\"");
		if(pos <= 0) return;
		str = str.substring(0, pos);
		String del = "M";
		pos = 0;
		Point2D.Double startL = null;
		while(pos >= 0){
			pos = str.indexOf(del);
			if(pos < 0) return;
			str = str.substring(pos + 1);
			pos = str.indexOf(" ");
			if(pos < 0) return;
			String a = str.substring(0, pos);
			str = str.substring(pos + 1);
			int x = Integer.parseInt(a,10);
			pos = str.indexOf(" ");
			if(pos < 0) {
				a = str;
				str = "";
			}
			else {
				a = str.substring(0, pos);
				str = str.substring(pos);
			}
			int y = Integer.parseInt(a, 10);
			Point2D.Double p = new Point2D.Double((double) x, (double) y);
			if(del.equals("M")) del = "L";
			else add(startL, p);
			startL = p;
		}
	}

	public Point2D.Double upscale(Point2D.Double p){
		float s = jpage.getScale();
		double x = p.getX() * s;
		double y = p.getY() * s;
		return new Point2D.Double(x, y);
	}

	public Point2D.Double dnscale(Point2D.Double p){
		float s = jpage.getScale();
		double x = p.getX() / s;
		double y = p.getY() / s;
		return new Point2D.Double(x, y);
	}

	public String getParm(String str, String parm){
		int pos = str.indexOf(parm);
		if(pos < 0) return null;
		str = str.substring(pos);
		pos = str.indexOf("\"");
		if (pos < 0) return null;
		str = str.substring(pos + 1);
		pos = str.indexOf("\"");
		if (pos < 0) return null;
		return str.substring(0, pos);
	}

	private void setClick(Jpage parent, Jtool jt, Point2D.Double click){
		jpage = parent;
		this.click = dnscale(click);
		float swidth = jt.getWidth() * cwidth/2;
		this.click = new Point2D.Double(this.click.getX() - swidth, this.click.getY() - swidth);
		jtool.copy(jt);	
		updateRectangle();
	}

	public Line2D.Double add(Point2D.Double startL, Point2D.Double endL){
		Line2D.Double dline = new Line2D.Double(dnscale(startL), dnscale(endL));
		lines.add(dline);
		updateRectangle();
		return dline;
	}

	public boolean below(int y){
		if(isPage) return false;
		y = (int)(y / jpage.getScale());
		if(y < ymin) return true;
		return false;
	}

	public boolean hit(Point2D.Double p){
		if(isPage) return false;
		p = dnscale(p);
		int x = (int) p.getX();
		int y = (int) p.getY();
		double rad = jtool.getWidth();
		if(click != null) rad = rad * cwidth;
		double s = (float) jpage.getScale();
		rad = (float) (rad + (8/s));
		int r = (int) rad;
		boolean test = true;
		if(x < xmin - r) test = false;
		if(x > xmax + r) test = false;
		if(y < ymin - r) test = false;
		if(y > ymax + r) test = false;
		return test;
	}

	public boolean overlaps(int a, int b, int w, int h){
		if(isPage) return false;
		double rad = jtool.getWidth();
		if(click != null) rad = rad * cwidth;
		double s = (float) jpage.getScale();
		rad = (float) (rad + (8/s));
		int r = (int) rad;
		if((a + w)/s < xmin - r) return false;
		if(a/s > xmax + r) return false;
		if((b + h)/s < ymin - r) return false;
		if(b/s > ymax + r) return false;
		return true;
	}

	public Rectangle getRectangle(){
		double s = (double) jpage.getScale();
		double rad = (s * jtool.getWidth())/ 2.0;
		if(click != null) rad = cwidth * rad;
		int x = (int) Math.floor((s * xmin) - rad);
		int y = (int) Math.floor((s * ymin) - rad);
		int w = (int) Math.ceil((s*(xmax - xmin)) + (2 * rad));
		int h = (int) Math.ceil((s*(ymax - ymin)) + (2 * rad));
		return (new Rectangle(x - 1,y - 1,w + 3,h + 3));
	}

	public void updateRectangle(){
		if(click == null){
			Line2D.Double dline = (Line2D.Double) lines.getLast();
			if(lines.size() == 1){
				xmin = (int) dline.getX1();
				xmax = xmin;
				ymin = (int) dline.getY1();
				ymax = ymin;
			}
			int x = (int) dline.getX2();
			int y = (int) dline.getY2();
			if(x < xmin) xmin = x;
			if(x > xmax) xmax = x;
			if(y < ymin) ymin = y;
			if(y > ymax) ymax = y;
		}
		else {
			float swidth = jtool.getWidth() * cwidth/2;
			xmin = (int) (click.getX() + swidth);
			xmax = xmin;
			ymin = (int) (click.getY() + swidth);
			ymax = ymin;
		}
	}				
	
	public void offset(Point2D.Double p){
		p = dnscale(p);
		for (int i = 0; i < lines.size(); i++){
			Line2D.Double dline = (Line2D.Double) lines.remove(i);	
			dline.setLine(dline.getX1() + p.getX(), dline.getY1() + p.getY(), dline.getX2() + p.getX(), dline.getY2() + p.getY());
			lines.add(i,dline);
		}
		if(click != null) click = new Point2D.Double(click.getX() + p.getX(), click.getY() + p.getY());
		xmin = (int)(xmin + p.getX());
		xmax = (int)(xmax + p.getX());
		ymin = (int)(ymin + p.getY());
		ymax = (int)(ymax + p.getY());
	}

	public boolean draw(Graphics2D g2, boolean print){
		if(isPage) return true;
		g2.setPaint(jtool.getColor());
		if(isHighlight){
			Color co = jtool.getColor();
			g2.setPaint(new Color(co.getRed(), co.getGreen(), co.getBlue(), jpage.getTransparency()));
		}
		float swidth = jtool.getWidth() * jpage.getScale();
		BasicStroke bs = new BasicStroke(swidth, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND);
		g2.setStroke(bs);
		if(click != null){
			Point2D.Double click = upscale(this.click);
			g2.fill(new Ellipse2D.Double(click.getX(), click.getY(), cwidth * swidth, cwidth * swidth));
			return false;
		}
		for (Iterator i = lines.iterator(); i.hasNext(); ){
			Line2D.Double dline = (Line2D.Double) i.next();
			dline = new Line2D.Double(upscale((Point2D.Double)dline.getP1()), upscale((Point2D.Double)dline.getP2()));
	    		g2.draw(dline);
		}
		return false;
	}
	
	public String save(){
		if(isPage) return "\n\n\n\n";
		String str = new String("");
		float swidth = jtool.getWidth();
		if(click != null){
			str = "<circle cx=\"" + (long)click.getX() + "\" cy=\"" + (long)click.getY() + "\" r=\"" + swidth + "\" fill=\"" + jtool.color + "\"/>\n";
			return str;
		}
		str = "<path d=\"";
		String sep = new String("M");
		for (Iterator i = lines.iterator(); i.hasNext(); ){
			Line2D.Double dline = (Line2D.Double) i.next();
			if(sep.equals("M")){
				str = str + sep + (long)dline.getX1() + " " + (long)dline.getY1();
				sep = "L";
			}
			str = str + " " + sep + (long)dline.getX2() + " " + (long)dline.getY2();
		
		}
		str = str + "\" stroke=\"" + jtool.color + "\" stroke-width=\"" + swidth + "\" fill=\"none\"/>\n";
		return str;
	}
}

class Jtext extends Jstroke{

	private Jpar jpar;

	public Point2D.Double corner;
	private int width;
	private int height = -1;

	public Jtext(Jpage parent, Point2D.Double corner, Jtool jt, int width, Jchar parms){
		jpage = parent;
		this.corner = corner;
		jtool = jt;
		this.width = width;
		height = 10;
		isPage = false;
		updateRectangle();
		jpar=new Jpar();
		jpar.setParms(parms);
		jpar.setWidth(width);
		isText = true;
	}

	public Jtext(Jpage parent, String str){
		jpage = parent;
		isPage = false;
		isText = true;
		make(str);
	}

	public String typeKey(String str){
		String oldXML = save();
		jpar.typeChar(str);
		height = jpar.height;
		updateRectangle();
		return oldXML;
	}

	public String setSelStyle(boolean bold, boolean italic, boolean underline, Float size, String font, String color){
		String oldXML = save();
		jpar.setSelStyle(bold, italic, underline, size, font, color);
		height = jpar.height;
		updateRectangle();
		return oldXML;
	}

	private void initRectangle(){
		if(height == -1){
			height = jpar.height;
			width = jpar.width;
		}
		updateRectangle();
	}

	public boolean hit(Point2D.Double p){
		p = dnscale(p);
		int x = (int) p.getX();
		int y = (int) p.getY();
		boolean test = true;
		if(x < xmin) test = false;
		if(x > xmax) test = false;
		if(y < ymin) test = false;
		if(y > ymax) test = false;
		return test;
	}

	public Rectangle getTextRectangle(){
		return getRectangle();
	}

	public Rectangle getRectangle(){
		double s = (double) jpage.getScale();
		int x = (int) Math.floor((s * xmin));
		int y = (int) Math.floor((s * ymin));
		int w = (int) Math.ceil((s*(xmax - xmin)));
		int h = (int) Math.ceil((s*(ymax - ymin)));
		return (new Rectangle(x - 1,y - 1,w + 3,h + 3));
	}

	public void updateRectangle(){
		xmin = (int) corner.getX();
		ymin = (int) corner.getY();
		xmax = xmin + width;
		ymax = ymin + height;
	}	
	
	public void offset(Point2D.Double p){
		p = dnscale(p);
		corner = new Point2D.Double(corner.getX() + p.getX(), corner.getY() + p.getY());
		updateRectangle();
	}

	public boolean draw(Graphics2D g2, boolean print){
		initRectangle();
		jpar.draw((int) corner.getX(), (int) corner.getY(), jpage.getScale(), g2, print);
		return false;
	}

	public void clearSel(){
		jpar.clearSel();
		height = jpar.height;
		updateRectangle();
	}
	
	public String save(){
		return jpar.getSVG((int)corner.getX(), (int) corner.getY());
	}

	public void make(String str){
		String a = super.getParm(str, "x=");
		if (a == null) return;
		int x = Integer.parseInt(a, 10);
		a = super.getParm(str, "y=");
		if (a == null) return;
		int y = Integer.parseInt(a, 10);
		corner = new Point2D.Double(x, y);
		jpar = new Jpar(str);
		width = jpar.width;
		updateRectangle();
	}
	
	public Point2D.Double hitSel(Point2D.Double p, float s){
		Point2D.Double pp = new Point2D.Double(p.getX()-(s*corner.getX()), p.getY()-(s*corner.getY()));
		jpar.hitSel(pp, s, false);
		jpar.findSel();
		return corner;
	}

	public Rectangle dragText(int x, int y, float s){
		Point2D.Double pp = new Point2D.Double(x-(s*corner.getX()), y-(s*corner.getY()));
		jpar.hitSel(pp, s, true);
		jpar.findSel();
		return getRectangle();
	}

	public Rectangle adv(int adv, int extend){
		jpar.advSel(adv, extend);
		jpar.findSel();
		return getRectangle();
	}
	public void clip(Toolkit toolkit){ 
		jpar.clipSel(toolkit);
	}
	public int getWidth(){
		return jpar.width;
	}
	public Rectangle setWidth(double x){
		Rectangle r = getRectangle();
		width = (int) (x - corner.getX());
		jpar.width = width;
		jpar.make();
		jpar.findSel();
		height = jpar.height;
		updateRectangle();
		return Jtool.maxR(r, getRectangle(), (int) (jpage.getScale() * 10));
	}
	String getDesc(){ return jpar.getDesc();}
	Jchar getCurParms(){ return jpar.getCurParms();}
	String getText(){ return jpar.getText();}
	String getHTML(){ return jpar.getHTML();}		
}

